#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QDebug>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    submenu(nullptr),
    newFile(new QAction(QIcon(":/catalogA/icons/new_file.png"),            tr("Create New File"))),
    openFile(new QAction(QIcon(":/catalogA/icons/open_file.png"),          tr("Open File"))),
    saveFile(new QAction(QIcon(":/catalogA/icons/save_file.png"),          tr("Save File"))),
    saveAllFiles(new QAction(QIcon(":/catalogA/icons/save_all_files.png"), tr("Save All Files"))),
    number(1),
    bug_t(true)
{
    ui->setupUi(this);
    setAcceptDrops(true);
    setWindowIcon(QIcon(":/catalogA/icons/icon.ico"));
    setWindowTitle("Text Editor");
    SetupTabWidget();
    SetupFileExplorer();
    SetupOpenDocs();
    CreateFile();
    SetupMenuBar();
    SetupToolBar();

    setCentralWidget(tabs);
    tabs->currentWidget()->setFocus();
    highlighter = new Highlighter(":/catalogA/highlights/languages.xml", this);

    connect(newFile,  SIGNAL(triggered()), SLOT(CreateFile()));
    newFile->setShortcut(Qt::CTRL + Qt::Key_N);
    connect(openFile,  SIGNAL(triggered()), SLOT(OpenFile()));
    openFile->setShortcut(Qt::CTRL + Qt::Key_O);
    connect(saveFile,  SIGNAL(triggered()), SLOT(SaveFile()));
    saveFile->setShortcut(Qt::CTRL + Qt::Key_S);
    connect(saveAllFiles,  SIGNAL(triggered()), SLOT(SaveAllFiles()));
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::dragEnterEvent(QDragEnterEvent* drag_event)
{
    if (drag_event->mimeData()->hasUrls())
        drag_event->acceptProposedAction();
}

void MainWindow::dropEvent(QDropEvent* drop_event)
{
    QList<QUrl> url_list = drop_event->mimeData()->urls();
    foreach (QUrl url, url_list) {
        OpenFile(url.url(QUrl::RemoveScheme));
    }
}

void MainWindow::SetupTabWidget()
{
    tabs = new QTabWidget(this);
    tabs->setMovable(true);
    tabs->setTabsClosable(true);
    tabs->setUsesScrollButtons(true);
    connect(tabs, SIGNAL(tabCloseRequested(int)), this, SLOT(CloseFile(int)));
    QToolButton* newT_button = new QToolButton(this);
    newT_button->setText("+");
    connect(newT_button, SIGNAL(clicked()), SLOT(CreateFile()));
    tabs->setCornerWidget(newT_button, Qt::TopLeftCorner);
}

void MainWindow::SetupMenuBar()
{
    QMenu* fileMenu = new QMenu(tr("File"), this);
    QMenu* editMenu = new QMenu(tr("Edit"), this);
    QMenu* viewMenu = new QMenu(tr("View"), this);

    fileMenu->addAction(newFile);
    fileMenu->addAction(openFile);  fileMenu->addSeparator();
    fileMenu->addAction(saveFile);
    fileMenu->addAction(tr("Save As..."), this, SLOT(SaveFileAs()),    Qt::SHIFT + Qt::CTRL + Qt::Key_S);
    fileMenu->addAction(saveAllFiles); fileMenu->addSeparator();
    fileMenu->addAction(tr("Close File"), this, SLOT(CloseFile()),     Qt::CTRL + Qt::Key_W);
    fileMenu->addAction(tr("Close All"),  this, SLOT(CloseAllFiles()), Qt::SHIFT + Qt::CTRL + Qt::Key_W); fileMenu->addSeparator();
    fileMenu->addAction(tr("Exit"),       this, SLOT(CloseWindow()),   Qt::CTRL + Qt::Key_Q);

    editMenu->addAction(tr("Cut"),        this, SLOT(Cut()),       Qt::CTRL + Qt::Key_X);
    editMenu->addAction(tr("Copy"),       this, SLOT(Copy()),      Qt::CTRL + Qt::Key_C);
    editMenu->addAction(tr("Paste"),      this, SLOT(Paste()),     Qt::CTRL + Qt::Key_V);
    editMenu->addAction(tr("Delete"),     this, SLOT(CLear()),     Qt::CTRL + Qt::Key_Backspace);
    editMenu->addAction(tr("Select All"), this, SLOT(SelectAll()), Qt::CTRL + Qt::Key_A);
    viewMenu->addAction(f_explorer->toggleViewAction());
    viewMenu->addAction(open_docs->toggleViewAction());

    ui->menubar->addMenu(fileMenu);
    ui->menubar->addMenu(editMenu);
    ui->menubar->addMenu(viewMenu);
}

void MainWindow::SetupToolBar()
{
    ui->mainToolBar->addAction(newFile);
    ui->mainToolBar->addAction(openFile);
    ui->mainToolBar->addAction(saveFile);
    ui->mainToolBar->addAction(saveAllFiles);
}

void MainWindow::SetupFileExplorer()
{
    f_sys_model = new QDirModel(this);
    f_sys_view  = new QTreeView(this);
    f_sys_view->setModel(f_sys_model);
    f_sys_view->setRootIndex(f_sys_model->index("/home"));
    f_sys_view->setColumnHidden(1, true);
    f_sys_view->setColumnHidden(2, true);
    f_sys_view->setColumnHidden(3, true);
    f_sys_view->setHeaderHidden(true);
    connect(f_sys_view, SIGNAL(doubleClicked(QModelIndex)), SLOT(OpenFile(QModelIndex)));

    f_explorer = new QDockWidget("File explorer", this);

    f_explorer->setWidget(f_sys_view);
    f_explorer->setFeatures(QDockWidget::DockWidgetClosable);
    f_explorer->hide();

    addDockWidget(Qt::LeftDockWidgetArea, f_explorer);
}

void MainWindow::SetupOpenDocs()
{
    open_docsW = new QListWidget(this);
    connect(tabs->tabBar(),      SIGNAL(currentChanged(int)),          SLOT(UpdateCurrentIndex(int)));
    connect(tabs->tabBar(),      SIGNAL(tabCloseRequested(int)),       SLOT(UpdateCurrentIndexOnDelete(int)));
    connect(tabs->tabBar(),      SIGNAL(tabMoved(int, int)),           SLOT(UpdateContains(int, int)));
    connect(tabs->tabBar(),      SIGNAL(tabMoved(int, int)),           SLOT(ChangeTabIndexInList(int, int)));
    connect(open_docsW,  SIGNAL(itemClicked(QListWidgetItem*)),SLOT(UpdateCurrentIndex(QListWidgetItem*)));
    connect(open_docsW,  SIGNAL(currentRowChanged(int)),       tabs, SLOT(setCurrentIndex(int)));

    open_docs  = new QDockWidget("Opened files", this);

    open_docs->setWidget(open_docsW);
    open_docs->setFeatures(QDockWidget::DockWidgetClosable);
    open_docs->hide();
    addDockWidget(Qt::RightDockWidgetArea, open_docs);
}

void MainWindow::CreateFile()
{
    Editor* new_text_edit = new Editor(this);
    int index = tabs->addTab(new_text_edit, "NoName" + QString::number(number));
    tabs->setCurrentIndex(index);
    tabs->setTabToolTip(index, "");
    tabs->setTabWhatsThis(index, "No changes");
    connect(new_text_edit, SIGNAL(textChanged()), SLOT(UpdateParam()));

    file_path_menu.push_back(QDir::current());
    file_name_menu.push_back("NoName" + QString::number(number));
    number++;

    QListWidgetItem* new_item = new QListWidgetItem;
    new_item->setText(tabs->tabText(index));
    new_item->setToolTip(tabs->tabToolTip(index));
    open_docsW->addItem(new_item);

    UpdateCurrentIndex(index);
}

void MainWindow::OpenFile()
{
    QString filepath = QFileDialog::getOpenFileName(this, "Choose file", "/home");
    if (filepath.isEmpty())
        return;
    OpenFile(filepath);
}

void MainWindow::OpenFile(const QString& filepath)
{
    for (int i = 0; i < tabs->count(); ++i)
        if (tabs->tabToolTip(i) == filepath)
        {
            QMessageBox::warning(this, tr("Duplicate"), tr("Cannot open the file twice"),
                                          QMessageBox::Ok);
            return;
        }

    QString filename = filepath.section("/",-1,-1);
    QFile file(filepath);

    if (file.open(QIODevice::ReadOnly))
    {
        Editor* temp_text = (Editor*)tabs->currentWidget();
        if (temp_text->document()->isEmpty() &&
            tabs->tabToolTip(tabs->currentIndex()) == "" &&
            tabs->tabText(tabs->currentIndex()) == "NoName")
        {

            DeleteTabFromList(tabs->currentIndex());
            delete tabs->widget(tabs->currentIndex());
        }

        file_path_menu.push_back(QFileInfo(file).path());
        file_name_menu.push_back(filename);

        Editor* new_text_edit = new Editor(this);
        QString content = file.readAll();
        file.close();
        new_text_edit->appendPlainText(content);
        int index = tabs->addTab(new_text_edit, filename);
        tabs->setCurrentIndex(index);
        tabs->setTabToolTip(index, filepath);
        tabs->setTabWhatsThis(index, "No changes");

        QListWidgetItem* new_item = new QListWidgetItem;
        new_item->setText(tabs->tabText(index));
        new_item->setToolTip(tabs->tabToolTip(index));
        open_docsW->addItem(new_item);

        QString file_extension = QFileInfo(filename).suffix();
        if (highlighter->setExtension(file_extension))
        {
            highlighter->setDocument(new_text_edit->document());
            highlighter->highlightBlock(new_text_edit->toPlainText());
        }

        tabs->setTabWhatsThis(index, "No changes");
        UpdateCurrentIndex(index);
        connect(new_text_edit, SIGNAL(textChanged()), SLOT(UpdateParam()));
        bug_t = false;
    }
    else
    {
        (new QErrorMessage(this))->showMessage("Can't open file");
        return;
    }
}

void MainWindow::SaveFile()
{
    if (tabs->tabToolTip(tabs->currentIndex()) == "")
    {
        SaveFileAs();
        return;
    }
    QString filepath = tabs->tabToolTip(tabs->currentIndex());
    QFile file(filepath);
    if (file.open(QIODevice::WriteOnly))
    {
        file.write(((Editor*)tabs->currentWidget())->document()->toPlainText().toUtf8());
        file.close();
        tabs->setTabWhatsThis(tabs->currentIndex(), "No changes");
    }
    else
    {
        (new QErrorMessage(this))->showMessage(tr("Can't save file"));
        return;
    }
}

void MainWindow::SaveFileAs()
{
    QString filename = tabs->tabText(tabs->currentIndex());
    filename.remove('*');
    QString filepath = QFileDialog::getSaveFileName(this, "Save " + filename, "/home/" + filename);
    if (filepath.isEmpty())
        return;
    if (QFileInfo(filepath).suffix().isEmpty())
        filepath.append(".txt");
    else
    {
        QFile file(filepath);
        if (file.open(QIODevice::WriteOnly))
        {
            file.write(((Editor*)tabs->currentWidget())->document()->toPlainText().toUtf8());
            file.close();
        }
        else
        {
            (new QErrorMessage(this))->showMessage(tr("Can't save file"));
            return;
        }
    }

    file_path_menu.replace(tabs->currentIndex(), QDir(QFileInfo(filepath).path()));
    file_name_menu.replace(tabs->currentIndex(), (QFileInfo(filepath).fileName()));

    filename = filepath.section("/",-1,-1);
    tabs->tabBar()->setTabText(tabs->currentIndex(), filename);
    tabs->tabBar()->setTabToolTip(tabs->currentIndex(), filepath);

    QString file_extension = QFileInfo(filename).suffix();
    if (highlighter->setExtension(file_extension))
    {
        highlighter->setDocument(((Editor*)tabs->currentWidget())->document());
        highlighter->highlightBlock(((Editor*)tabs->currentWidget())->toPlainText());
    }
    tabs->setTabWhatsThis(tabs->currentIndex(), "No changes");
    MainWindow::SaveFile();
}

void MainWindow::SaveAllFiles()
{
    int current_index = tabs->currentIndex();
    for (int i = 0; i < tabs->count(); ++i)
    {
        tabs->setCurrentIndex(i);
        if (tabs->tabWhatsThis(tabs->currentIndex()) != "No changes")
            SaveFile();
    }
    tabs->setCurrentIndex(current_index);
}

void MainWindow::CloseFile(int idx_)
{
    if (tabs->tabWhatsThis(tabs->currentIndex()) != "No changes")
    {
        QMessageBox::StandardButton reply;
        reply = QMessageBox::question(this, tr("Saving changes"), tr("Save changes before closing?"),
                                      QMessageBox::Yes | QMessageBox::No);
        if (reply == QMessageBox::Yes)
        {
            SaveFile();
        }
    }
    file_path_menu.remove(idx_);
    file_name_menu.removeAt(idx_);
    delete tabs->widget(idx_);
    DeleteTabFromList(idx_);
    if (!tabs->count())
        CreateFile();
    tabs->currentWidget()->setFocus();
}

void MainWindow::CloseFile()
{
    CloseFile(tabs->currentIndex());
}

void MainWindow::CloseAllFiles()
{
    bool checker = false;
    for (int i = 0; i < tabs->count(); ++i)
    {
        if (tabs->tabWhatsThis(i) != "No changs")
        {
            checker = true;
            break;
        }
    }
    if (checker)
    {
        QMessageBox::StandardButton reply = QMessageBox::question(
            this, tr("Save all changes"), tr("Save all changes before closing?"),
            QMessageBox::Yes | QMessageBox::No);
        if (reply == QMessageBox::Yes)
            SaveAllFiles();
    }

    file_path_menu.clear();
    file_name_menu.clear();
    number = 1;

    while (tabs->count() > 0)
        delete tabs->widget(0);
    open_docsW->clear();
    CreateFile();
    tabs->currentWidget()->setFocus();
}

void MainWindow::CloseWindow()
{
    if(submenu)
        delete submenu;
    submenu = new ExitMenu(this);
    submenu->setWindowTitle(tr("Save changes?"));

    QVector<QDir> temp_path;
    QStringList temp_file_names;

    for(int i = 0; i < tabs->count(); i++)
    {
        if(tabs->tabText(i).contains('*'))
        {
            temp_path.push_back(file_path_menu[i]);
            temp_file_names.push_back(file_name_menu[i]);
        }
    }

    if(!temp_path.isEmpty())
    {
        submenu->Fill(temp_path, temp_file_names);
        int event = submenu->exec();
        if(event == QDialog::Accepted)
            SaveAllFiles();
        if(event != -1)
            QApplication::quit();
    }
    else
        QApplication::quit();
}

void MainWindow::closeEvent(QCloseEvent* event)
{
    event->ignore();
    CloseWindow();
}

void::MainWindow::UpdateParam()
{
    QString file = tabs->tabBar()->tabText(tabs->currentIndex());
    QString file_extension = QFileInfo(file).suffix();
    if (!file_extension.isEmpty())
    {
        if (highlighter->setExtension(file_extension))
        {
            if(bug_t)
            {
                tabs->setTabWhatsThis(tabs->currentIndex(), "Changed");
                tabs->setTabText(tabs->currentIndex(), tabs->tabText(tabs->currentIndex()) + "*");
                disconnect(sender(), SIGNAL(textChanged()), this, SLOT(UpdateParam()));
            }
            else
                bug_t = true;
            return;
        }
    }

    tabs->setTabWhatsThis(tabs->currentIndex(), "Changed");
    tabs->setTabText(tabs->currentIndex(), tabs->tabText(tabs->currentIndex()) + "*");
    disconnect(sender(), SIGNAL(textChanged()), this, SLOT(UpdateParam()));
}

void MainWindow::ChangeTabIndexInList(int old_index, int new_index)
{
    QListWidgetItem* first_item  = open_docsW->takeItem(old_index);
    open_docsW->insertItem(new_index, first_item);
}

void MainWindow::OpenFile(QModelIndex file_index)
{
    if (!f_sys_model->isDir(file_index))
        OpenFile(f_sys_model->filePath(file_index));
}

void MainWindow::DeleteTabFromList(int index)
{
    QListWidgetItem* temp_item = open_docsW->takeItem(index);
    delete temp_item;
}

void MainWindow::UpdateCurrentIndexOnDelete(int)
{
    open_docsW->setCurrentRow(open_docsW->count() - 1);
}

void MainWindow::UpdateCurrentIndex(QListWidgetItem* current_item)
{
    int index = current_item->listWidget()->row(current_item);
    tabs->setCurrentIndex(index);
}

void MainWindow::UpdateCurrentIndex(int new_selection_index)
{
    open_docsW->setCurrentRow(new_selection_index);

    QString file = tabs->tabBar()->tabText(new_selection_index);
    QString file_extension = QFileInfo(file).suffix();
    if (!file_extension.isEmpty())
    {
        if (highlighter->setExtension(file_extension))
        {
            highlighter->setDocument(((Editor*)tabs->currentWidget())->document());
            highlighter->highlightBlock(((Editor*)tabs->currentWidget())->toPlainText());
        }
    }
}

void MainWindow::Copy()
{
    ((Editor*)tabs->currentWidget())->copy();
}

void MainWindow::Cut()
{
    ((Editor*)tabs->currentWidget())->cut();
}

void MainWindow::Paste()
{
    ((Editor*)tabs->currentWidget())->paste();
}

void MainWindow::SelectAll()
{
    ((Editor*)tabs->currentWidget())->selectAll();
}

void MainWindow::CLear()
{
    ((Editor*)tabs->currentWidget())->clear();
}

void MainWindow::UpdateContains(int from, int to)
{
    QDir temp = file_path_menu.at(from);
    file_path_menu.replace(from, file_path_menu.at(to));
    file_path_menu.replace(to, temp);
    QString temp_file_name = file_name_menu.at(from);
    file_name_menu.replace(from, file_name_menu.at(to));
    file_name_menu.replace(to, temp_file_name);
}
